@@ -88,7 +88,7 @@ Dtype dtype_of_buffer(Py_buffer *view)
     // Currently, we only understand native endianness and alignment.
     if (*fmt == '@') fmt++;
 
-    if (strchr("cbB?hHiIlL", *fmt)) {
+    if (strchr("cbB?hHiIlLqQnN", *fmt)) {
         dtype = LONG;
         fmt++;
     } else if (strchr("fdg", *fmt)) {
@@ -103,7 +103,7 @@ Dtype dtype_of_buffer(Py_buffer *view)
     }
 
     // Right now, no composite data structures are supported; if we found a
-    // single supported data type, we should be a the end of the string.
+    // single supported data type, we should be at the end of the string.
     if (*fmt != '\0') return NONE;
 
     return dtype;
@@ -365,9 +365,9 @@ PyObject *(*readin_scalar_into_new_dtable[])(PyObject*, bool, int) =
 
 int examine_buffer(PyObject *in, Py_buffer *view, Dtype *dtype)
 {
+    if (!PyObject_CheckBuffer(in)) return -1;
     Dtype dt = NONE;
     memset(view, 0, sizeof(Py_buffer));
-    if (!PyObject_CheckBuffer(in)) return -1;
 
     // I don't know if the following makes much sense: I try to get the buffer
     // using less and less demanding flags. NumPy does the same.
@@ -458,6 +458,10 @@ T (*get_buffer_converter(Py_buffer *view))(const void *)
         return number_from_ptr<T, long long>;
     case 'Q':
         return number_from_ptr<T, unsigned long long>;
+    case 'n':
+        return number_from_ptr<T, ssize_t>;
+    case 'N':
+        return number_from_ptr<T, size_t>;
     case 'f':
         return number_from_ptr<T, float>;
     case 'd':
@@ -850,8 +854,19 @@ Py_hash_t hash(Complex x)
     return hash(x.real()) + HASH_IMAG * hash(x.imag());
 }
 
-// This routine calculates the hash of a multi-dimensional array.  The hash is
-// equal to that of an arrangement of nested tuples equivalent to the array.
+// The following routine calculates the hash of a multi-dimensional array.  The
+// hash is equal to that of an arrangement of nested tuples equivalent to the
+// array.
+//
+// It exists in two versions because Python's tuplehash has changed in Python
+// 3.8 with the following motivation: "The hash function for tuples is now
+// based on xxHash which gives better collision results on (formerly)
+// pathological cases. Additionally, on 64-bit systems it improves tuple hashes
+// in general."
+
+#if (PY_MAJOR_VERSION < 3 || PY_MINOR_VERSION < 8) && PY_MAJOR_VERSION < 4
+
+// Version for Python < 3.8
 template <typename T>
 Py_hash_t hash(PyObject *obj)
 {
@@ -875,16 +890,20 @@ Py_hash_t hash(PyObject *obj)
         if (i[d]) {
             --i[d];
             if (d == ndim) {
+                // Innermost loop body.
                 r[d] = (r[d] ^ hash(*p++)) * mult[d];
                 mult[d] += mul_addend + 2 * i[d];
             } else {
+                // Entering a loop.
                 ++d;
                 i[d] = shape[d];
                 mult[d] = mult_init;
                 r[d] = r_init;
             }
         } else {
+            // Exiting a loop.
             if (d == 0) {
+                // Exiting the outermost loop.
                 Py_uhash_t r_next = r[0] + r_addend;
                 return r_next == Py_uhash_t(-1) ? -2 : r_next;
             }
@@ -897,6 +916,77 @@ Py_hash_t hash(PyObject *obj)
     }
 }
 
+#else
+
+#if SIZEOF_PY_UHASH_T > 4
+
+const Py_uhash_t _hash_init = 2870177450012600261U;
+
+inline void _hash_inner_loop(Py_uhash_t &acc, Py_uhash_t lane)
+{
+    acc += lane * 14029467366897019727U;
+    acc = ((acc << 31) | (acc >> 33)); // Rotate left 31 bits.
+    acc *= 11400714785074694791U;
+}
+
+#else
+
+const Py_uhash_t _hash_init = 374761393U;
+
+inline void _hash_inner_loop(Py_uhash_t &acc, Py_uhash_t lane)
+{
+    acc += lane * 2246822519U;
+    acc = ((acc << 13) | (acc >> 19)); // Rotate left 13 bits.
+    acc *= 2654435761U;
+}
+
+#endif
+
+inline Py_uhash_t _hash_loop_end(Py_uhash_t acc, Py_uhash_t len)
+{
+    acc += len ^ (_hash_init ^ 3527539UL);
+    if (acc == Py_uhash_t(-1)) return 1546275796;
+    return acc;
+}
+
+// Version for Python >= 3.8
+template <typename T>
+Py_hash_t hash(PyObject *obj)
+{
+    int ndim;
+    size_t *shape;
+    Array<T> *self = reinterpret_cast<Array<T> *>(obj);
+    self->ndim_shape(&ndim, &shape);
+    T *p = self->data();
+    if (ndim == 0) return hash(*p);
+
+    Py_ssize_t i[max_ndim];
+    Py_uhash_t acc[max_ndim];
+    --ndim;                     // For convenience.
+    int d = 0;
+    i[0] = shape[0];
+    acc[0] = _hash_init;
+    // The following is equivalent to 'ndim' (the original value) nested loops.
+    while (true) {
+        if (i[d]) {
+            --i[d];
+            if (d == ndim) {
+                _hash_inner_loop(acc[d], hash(*p++));
+            } else {
+                ++d;
+                i[d] = shape[d];
+                acc[d] = _hash_init;
+            }
+        } else {
+            if (d == 0) return _hash_loop_end(acc[0], shape[0]);
+            --d;
+            _hash_inner_loop(acc[d], _hash_loop_end(acc[d+1], shape[d+1]));
+        }
+    }
+}
+
+#endif
+
 template <typename T>
 bool compare_scalar(const int op, const T a, const T b) {
     switch(op){
